// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file CouplingProblemStat.h */

#ifndef AMDIS_COUPLING_PROBLEM_STAT_H
#define AMDIS_COUPLING_PROBLEM_STAT_H

#include <vector>
#include <set>
#include <list>
#include "AMDiS_fwd.h"
#include "ProblemStat.h"
#include "Initfile.h"
#include <boost/lexical_cast.hpp>

namespace AMDiS {

  using namespace std;


  /** \brief
   * This class defines a coupled stationary problem definition in sequential
   * computations. 
   */
  class CouplingProblemStat
  {
  public:
    /// Constructor
    CouplingProblemStat(std::string name_)
      : name(name_),
        nComponents(0),
        nMeshes(0),
        refinementManager(NULL),
        coarseningManager(NULL)
    {}

    /// Destructor
    virtual ~CouplingProblemStat() {}

    /// add problem by number
    virtual void addProblem(ProblemStat* prob)
    {
      problems.push_back(prob);
    };

    /// Initialisation of the problem.
    virtual void initialize(Flag initFlag,
                            CouplingProblemStat *adoptProblem = NULL,
                            Flag adoptFlag = INIT_NOTHING)
    {

      // create one refinement-/coarseningmanager for all problems
      if (refinementManager != NULL && coarseningManager != NULL) { 
        WARNING("refinement-/coarseningmanager already created\n");
      } else {
        if (!adoptProblem)
          createRefCoarseManager();
        else {
          refinementManager = adoptProblem->refinementManager;
          coarseningManager = adoptProblem->coarseningManager;
        }
      }
  
      if (refinementManager == NULL || coarseningManager == NULL) 
        WARNING("no refinement-/coarseningmanager created\n");

      // create Meshes and FeSpaces

      // one global macro-mesh for all problems
      string meshName("");
      Parameters::get(name + "->mesh", meshName);
      TEST_EXIT(meshName != "")("No mesh name specified for \"%s->mesh\"!\n", 
                                name.c_str());

      // all problems must have the same dimension (?)
      int dim = 0;
      Parameters::get(name + "->dim", dim);
      TEST_EXIT(dim)("No problem dimension specified for \"%s->dim\"!\n",
                    name.c_str());

      map<int, Mesh*> meshForRefinementSet;
      vector< std::set<Mesh*> > meshesForProblems(problems.size());
      map<pair<Mesh*, int>, FiniteElemSpace*> feSpaceMap;
      for (size_t i = 0; i < problems.size(); ++i) {
        TEST_EXIT(problems[i])("problem[%d] does not exist!\n",i);
        nComponents += problems[i]->getNumComponents();
        for (size_t j = 0; j < problems[i]->getNumComponents(); j++) {
          // mesh
          int refSet = -1;
          Parameters::get(problems[i]->getName() + "->" +
                          "refinement set[" + boost::lexical_cast<string>(j) + "]", 
                          refSet);
          if (refSet < 0)
            refSet = 0;
          
          if (meshForRefinementSet.find(refSet) == meshForRefinementSet.end()) {
            Mesh *newMesh = new Mesh(meshName, dim);
            meshForRefinementSet[refSet] = newMesh;
            meshes.push_back(newMesh);
            meshesForProblems[i].insert(newMesh);
            nMeshes++;
          } else
            meshesForProblems[i].insert(meshForRefinementSet[refSet]);

          problems[i]->setComponentMesh(j, meshForRefinementSet[refSet]);
            
          // feSpace
          int degree = 1;
          Parameters::get(problems[i]->getName() + "->polynomial degree[" + 
                          boost::lexical_cast<string>(j) + "]", degree);
    
          if (feSpaceMap[pair<Mesh*, int>(meshForRefinementSet[refSet], degree)] == NULL) {
            stringstream s;
            s << problems[i]->getName() << "->feSpace[" << j << "]";
    
            FiniteElemSpace *newFeSpace = 
              FiniteElemSpace::provideFeSpace(NULL, Lagrange::getLagrange(dim, degree),
                                              meshForRefinementSet[refSet], s.str());
            feSpaceMap[pair<Mesh*, int>(meshForRefinementSet[refSet], degree)] = newFeSpace;
            feSpaces.push_back(newFeSpace);
          }
    
//           problems[i].setComponentSpace(i, feSpaceMap[pair<Mesh*, int>(meshForRefinementSet[refSet], degree)]);
        }
      }

      for (size_t i = 0; i < problems.size(); ++i) {
        vector<Mesh*> problemMeshes(meshesForProblems[i].begin(), meshesForProblems[i].end());
        problems[i]->setMeshes(problemMeshes);
        problems[i]->setRefinementManager(refinementManager);
        problems[i]->setCoarseningManager(coarseningManager);
        problems[i]->initialize(INIT_ALL);
      }
    }

    void createRefCoarseManager() 
    {
      FUNCNAME("ProblemStat::createRefCoarseManager()");
  
      int dim = 0;
      Parameters::get(name + "->dim", dim);
      TEST_EXIT(dim)("No problem dimension specified for \"%s->dim\"!\n",
                    name.c_str());

      switch (dim) {
      case 1:
        coarseningManager = new CoarseningManager1d();
        refinementManager = new RefinementManager1d();
        break;
      case 2:
        coarseningManager = new CoarseningManager2d();
        refinementManager = new RefinementManager2d();
        break;
      case 3:
        coarseningManager = new CoarseningManager3d();
        refinementManager = new RefinementManager3d();
        break;
      default:
        ERROR_EXIT("invalid dim!\n");
      }
    }

    /// Returns number of managed problems
    virtual int getNumProblems() 
    { 
      return problems.size(); 
    }

    /// Implementation of ProblemStatBase::getNumComponents()
    virtual int getNumComponents() 
    { 
      return nComponents; 
    }

    /** \brief
     * Returns the problem with the given number. If only one problem
     * is managed by this master problem, the number hasn't to be given.
     */
    virtual ProblemStat *getProblem(int number = 0) 
    { 
      return problems[number]; 
    }

    /// Returns \ref meshes[i]
    inline Mesh* getMesh(int number = 0) 
    {
      return meshes[number]; 
    }

    /// Returns \ref meshes
    inline vector<Mesh*> getMeshes() 
    {
      return meshes; 
    }

    /// Returns \ref refinementManager.
    inline RefinementManager* getRefinementManager(int comp = 0) 
    { 
      return refinementManager; 
    }

    /// Returns \ref refinementManager.
    inline CoarseningManager* getCoarseningManager(int comp = 0) 
    { 
      return coarseningManager; 
    }

    /// Returns the name of the problem
    inline virtual string getName() 
    { 
      return name; 
    }

  protected:
    
    /// Name of this problem.
    string name;

    /// Number of problem components
    int nComponents;

    /** \brief
     * Number of problem meshes. If all components are defined on the same mesh,
     * this number is 1. Otherwise, this variable is the number of different meshes
     * within the problem.
     */
    int nMeshes;

    /// FE spaces of this problem.
    vector<FiniteElemSpace*> feSpaces;

    /// Meshes of this problem.
    vector<Mesh*> meshes;

    /** \brief
     * All actions of mesh refinement are performed by refinementManager.
     * If new refinement algorithms should be realized, one has to override
     * RefinementManager and give one instance of it to AdaptStationary.
     */
    RefinementManager *refinementManager;

    /** \brief
     * All actions of mesh coarsening are performed by coarseningManager.
     * If new coarsening algorithms should be realized, one has to override
     * CoarseningManager and give one instance of it to AdaptStationary.
     */
    CoarseningManager *coarseningManager;

    vector<ProblemStat*> problems;
  };
}

#endif
